"""OneToOne MapperProperty and Python Property implementations."""

from sqlalchemy.orm.interfaces import PropComparator, MapperProperty
from sqlalchemy.orm.properties import PropertyLoader
from sqlalchemy.orm.session import register_attribute

__all__ = [
    'OneToOneProxy', 'OneToOneMapperProperty', 
    'OneToOnePropertyLoader', 'one_to_one'
]

class OneToOneProxy(object):
    """Property that proxies attribute access to columns on a related object."""    
    def __init__(self, key, remote_class, remote_property_name):
        self.key = key
        self.remote_class = remote_class
        self.remote_property_name = remote_property_name
    
    def __get__(self, obj, type=None):
        if obj is None:
            return self
        value = None
        related_obj = getattr(obj, self.key, None)
        if related_obj:
            value = getattr(related_obj, self.remote_property_name, None)
        return value
    
    def __set__(self, obj, value):
        related_obj = getattr(obj, self.key, None)
        if related_obj is None:
            related_obj = self.remote_class()
            setattr(obj, self.key, related_obj)
        setattr(related_obj, self.remote_property_name, value)


class OneToOneMapperProperty(MapperProperty):
    """MapperProperty that proxies access to one column on a related object.
    
    This is a type of `MapperProperty` that proxies to the mapper's class
    property of the same name and allows queries to reference this property.
    The property's queries will be modified to contain a join condition
    from the parent mapper to the mapper used in the relation() given by the
    `one_to_one_relation` argument.
    
    This class is meant to be used with `OneToOnePropertyLoader` for the sole
    purpose of allowing proxied columns to be queried through the parent
    mapper.
    
    """
    def __init__(self, proxy_property, relation):
        """Tell the people what this does!"""
        self.proxy_property = proxy_property
        self.remote_property_name = self.proxy_property.remote_property_name
        self.relation = relation
        self.comparator = OneToOneMapperProperty.Comparator(self, None)
    
    def do_init(self):
        """Automatically called after key and parent are determined"""
        # Determine the table column this property is proxying to
        self.column = self.relation.mapper.get_property(
            self.remote_property_name
        ).columns[0]
        # Let the SQLAlchemy UOW know about this property
        register_attribute(self.parent.class_, self.key,
            uselist=False,
            proxy_property=self.proxy_property,
            useobject=False,
            comparator=self.comparator
        )
    
    def create_row_processor(self, *args, **kwargs):
        return (None, None)
    
    class Comparator(PropComparator):
        """
        This property comparator adds an extra condition to all comparison
        operations, so that the resulting clause will include the join
        condition between self.prop's parent mapper and
        self.prop.one_to_one_relation.
        
        This class is meant to be used with OneToOneMapperProperty for the
        sole purpose of ensuring that the correct comparison clause is
        generated when querying proxied columns.
        
        """
        def operate(self, op, *other, **kwargs): 
            join_clause = self.prop.relation.primaryjoin # ex. book.author_id = author.author_id
            op_clause = op(self.prop.column, *other, **kwargs) # ex. author.name == 'Maya Angelou'
            return join_clause & op_clause # ex. book.author_id = author.author_id AND author.name == 'Maya Angelou'
        
        def clause_element(self):
            return self.prop.column
        
        def reverse_operate(self, op, other, **kwargs):
            join_clause = self.prop.relation.primaryjoin
            column = self.prop.column
            op_clause = op(column._bind_param(other), column, **kwargs)
            return join_clause & op_clause


class OneToOnePropertyLoader(PropertyLoader):
    """PropertyLoader that proxies access to all columns on a related object.
    
    Because this PropertyLoader adds other properties, it must be added
    after the mapper's ``properties`` mapping is already defined, 
    meaning it must be added using ``my_mapper.add_property`` after the
    mapper has already defined.
    """
    def __init__(self, *args, **kwargs):
        kwargs['uselist'] = False
        super(OneToOnePropertyLoader, self).__init__(*args, **kwargs)
    
    def do_init(self):
        super(OneToOnePropertyLoader, self).do_init()
        # It's likely that no mappers have been compiled yet, so the 
        # properties in self.mapper.iterate_properties won't know their key...
        # SO we have to do something really unpolite to get the list of
        # properties.  Hopefully Alchemy will provide a better way soon...
        for prop_key, prop in self.mapper._Mapper__props.iteritems():
            # Skip foreign keys, they are already mapped
            if prop.columns[0] in self.remote_side:
                continue
            # Create the Python Property and add it to the mapped class
            proxy_property = OneToOneProxy(self.key, self.argument, prop_key)
            setattr(self.parent.class_, prop_key, proxy_property)
            
            # Create the MapperProperty so the proxied column will be query-able
            mapper_property = OneToOneMapperProperty(proxy_property, self)
            
            # Add the MapperProperty to the mapper.  Ideally this would be a
            # call to `self.parent.add_property`, but the mapper is in the
            # middle of compiling, meaning `add_property` doesn't work yet.
            # The `_compile_property` call below forces the added property
            # to initialize.
            self.parent._init_properties[prop_key] = mapper_property
            self.parent._compile_property(prop_key, mapper_property, True)

            
def one_to_one(*args, **kwargs):
    return OneToOnePropertyLoader(*args, **kwargs)


"""Test suite for python examples embedded in the documentation"""
import os
import doctest

def test_suite():
    # Locate the doc/src directory relative to the current file.
    test_dir = os.path.dirname(os.path.abspath(__file__))
    brainfreeze_dir = os.path.dirname(test_dir)
    doc_dir = os.path.join(brainfreeze_dir, 'doc', 'src')
    
    # Walk the doc/src tree and locate all the .txt files
    doc_files = []
    for root, dirs, files in os.walk(doc_dir):
        for name in files:
            if name.endswith('.txt'):
                path = os.path.join(root, name)
                doc_files.append(path)

    kw = dict(module_relative=False)
    return doctest.DocFileSuite(*doc_files, **kw)

if __name__ == '__main__':
    import sys
    import unittest
    if sys.version_info[:2] < (2, 4):
        print "Python 2.4 or later required for tests (%d.%d detected)." % \
            sys.version_info[:2]
        sys.exit(-1)
    unittest.TextTestRunner().run(test_suite())
    




#!python
"""Setuptools-based installation and packaging module"""

# Ensure setuptools is available (download & install if necessary)
import ez_setup
ez_setup.use_setuptools()
from setuptools import setup, find_packages

# Get package information (and googlecode_upload Command class)
import os
execfile(os.path.join("brainfreeze", "release", "__init__.py"))

# Check to make sure required python is available
def check_python():
    if sys.version_info < (2,3):
        print "Python 2.3 or higher is required to run BrainFreeze."
        sys.exit(1)

# Load long description from README file
readme = os.path.normpath(__file__ + "/../README")
f = open(readme)
LONG_DESCRIPTION = f.read()

# Implement a Setuptools Command to generate documentation
class makedocs(Command):
    """Setuptools Command to generate html and wiki documentation"""
    description = "generate BrainFreeze documentation"
    
    user_options = []
    
    def initialize_options(self):
        pass
        
    def finalize_options(self):
        pass

    def run(self):
        # Locate the doc/src directory relative to the current file.
        main_dir = os.path.dirname(os.path.abspath(__file__))
        doc_dir = os.path.join(main_dir, 'brainfreeze', 'doc', 'src')
        wiki_dir = os.path.join(os.path.dirname(main_dir), 'wiki')
        # Walk the doc/src tree and locate all the .txt files
        doc_files = []
        for root, dirs, files in os.walk(doc_dir):
            for name in files:
                if name.endswith('.txt'):
                    path = os.path.join(root, name)
                    doc_files.append(path)
        # Write all the .wiki files!
        from brainfreeze.doc.build import rst2wiki
        for path in doc_files:
            f = open(path)
            source = f.read()
            f.close()
            output = rst2wiki.publish_string(
                source, writer=rst2wiki.WikiWriter()
            )
            wiki_name = os.path.basename(path)[:-len('.txt')] + '.wiki'
            wiki_path = os.path.join(wiki_dir, wiki_name)
            wiki_file = open(wiki_path, 'w')
            wiki_file.write(output)

# Let setuptools work its magic!
setup(
    name=NAME,
    version=VERSION,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    url=URL,
    download_url=DOWNLOAD_URL,
	packages=find_packages(), # find python packages in this dir
	include_package_data=True, # include more than just the code
	cmdclass={
	    'googlecode_upload': googlecode_upload,
	    'makedocs': makedocs 
	},
	keywords=["brainfreeze", "sqlalchemy", "orm"],
    test_suite="brainfreeze.test.test_docs.test_suite",
    classifiers=[
         "Intended Audience :: Developers",
         "License :: OSI Approved :: BSD License",
         "Operating System :: OS Independent",
         "Programming Language :: Python",
         "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)


